#!/usr/bin/env python3
# Copyright 2020 Cole White
#                Filippo Giunchedi
#                Wikimedia Foundation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import datetime
import json
import logging
import os
import re
import subprocess
import sys
import threading
import time
import urllib
from dataclasses import dataclass
from typing import Dict, List, Optional, io

import urllib3.exceptions
import yaml
from prometheus_client import start_http_server
from prometheus_client.core import Counter, Gauge, Summary

import prometheus_icinga_exporter.alertmanager_client as am_client

LOG = logging.getLogger(__name__)

AM_REQUEST_TIMEOUT = 2
# Fraction of interval to dedicate to retrying 'status command' failures
STATUS_COMMAND_RETRY_FACTOR = 0.5

METRIC_NAMESPACE = "icinga_am"


class UnknownLabel(Exception):
    """Risen when there's an unknown label specified on the team labels config."""


@dataclass(frozen=True)
class TeamRegexes:
    instance: List[re.Pattern]
    alertname: List[re.Pattern]

    @classmethod
    def from_dict(cls, raw_dict: Dict[str, List[str]]) -> "TeamRegexes":
        raw_dict_copy = raw_dict.copy()
        instance_patterns = [
            re.compile(regex) for regex in raw_dict_copy.pop("instance", [])
        ]
        alertname_patterns = [
            re.compile(regex) for regex in raw_dict_copy.pop("alertname", [])
        ]
        if raw_dict_copy:
            raise UnknownLabel(
                f"Unknown labels {raw_dict_copy.keys()} found for config {raw_dict}."
            )

        return cls(
            instance=instance_patterns,
            alertname=alertname_patterns,
        )


@dataclass(frozen=True)
class TeamTagMatcher:
    # Storing compiled regexes makes searches faster
    teams_regexes: Dict[str, TeamRegexes]

    @classmethod
    def from_team_regexes_file(cls, team_regexes_file: io.TextIO) -> "TeamTagMatcher":
        try:
            raw_team_regexes = yaml.safe_load(team_regexes_file)
        except Exception as error:
            raise Exception(
                f"Unable to read team regexes file, got error: {error}"
            ) from error

        teams_regexes = {
            team_name: TeamRegexes.from_dict(raw_dict=regexes_dict)
            for team_name, regexes_dict in raw_team_regexes.items()
        }

        return cls(teams_regexes=teams_regexes)

    def get_matching_team(self, alertname: str, instance: str) -> Optional[str]:
        """Currently matching the first team only."""
        for team_name, team_regexes in self.teams_regexes.items():
            if any(regex.match(alertname) for regex in team_regexes.alertname):
                return team_name

            if any(regex.match(instance) for regex in team_regexes.instance):
                return team_name

        return None


class PrometheusIcingaAM:
    """Prometheus Icinga Alertmanager client"""

    parse_duration = Summary(
        "parse_duration_seconds",
        "Icinga AM parsing duration",
        (),
        namespace=METRIC_NAMESPACE,
    )

    metrics = {
        "last_parse_success_epoch": Gauge(
            "last_parse_success_epoch",
            "UNIX timestamp of last successful parse",
            (),
            namespace=METRIC_NAMESPACE,
        ),
        "parse_errors_total": Counter(
            "parse_errors_total",
            "Total number of errors encountered while acquiring and parsing status information",
            (),
            namespace=METRIC_NAMESPACE,
        ),
        "send_errors_total": Counter(
            "send_errors_total",
            "Total number of errors encountered while sending alerts",
            (),
            namespace=METRIC_NAMESPACE,
        ),
        "send_completed_total": Counter(
            "send_completed_total",
            "Total number of completed calls to Alertmanager",
            (),
            namespace=METRIC_NAMESPACE,
        ),
    }

    def __init__(
        self,
        alertmanagers,
        query_string,
        status_cgi_path,
        interval=10,
        notes_map=None,
        team_tag_matcher: Optional[TeamTagMatcher] = None,
        extinfo_url: Optional[str] = None,
    ):
        self.alertmanagers = alertmanagers
        self.am_clients = self._build_am_clients(alertmanagers)
        self.interval = interval
        self.status_cgi_path = status_cgi_path
        self.query_string = query_string
        self.notes_map = notes_map
        self.alerts = None
        self.team_tag_matcher = team_tag_matcher
        self.extinfo_url = extinfo_url

    def _build_am_clients(self, alertmanagers):
        """Build an am_client.AlertApi for each alertmanager."""

        res = []
        for address in alertmanagers:
            parsed = urllib.parse.urlparse(address, scheme="http")
            endpoint = parsed.geturl()
            if not parsed.path:
                endpoint += "/api/v2"

            api_client = am_client.ApiClient(am_client.Configuration(host=endpoint))
            LOG.debug("Build AM client for %r", endpoint)
            res.append(am_client.AlertApi(api_client))

        return res

    def start(self):
        """Start a thread to periodically parse problems and send alerts to AM"""

        def do_every(period, f, *args, **kwargs):
            def g_tick():
                t = time.time()
                while True:
                    t += period
                    yield max(t - time.time(), 0)

            g = g_tick()
            while True:
                time.sleep(next(g))
                f(*args, **kwargs)

        def _run():
            try:
                self._send_alerts()
            except Exception as e:
                LOG.exception(e)
                os._exit(-1)

        t = threading.Thread(target=do_every, args=(self.interval, _run))
        t.daemon = True
        t.start()
        return t

    def _parse_cgi_json(self, stdout):
        try:
            # Extract HTTP body from CGI's output
            status_json = stdout[stdout.index(b"\r\n\r\n") + 2 :]
            return json.loads(status_json)
        except (ValueError, json.decoder.JSONDecodeError):
            LOG.exception(
                "Failed to parse JSON from CGI HTTP body: %r...", stdout[:255]
            )
        return None

    def _json_from_status_cmd(self):
        env = {"REQUEST_METHOD": "GET", "QUERY_STRING": self.query_string}

        start_t = time.time()

        while time.time() < start_t + (self.interval * STATUS_COMMAND_RETRY_FACTOR):
            LOG.debug("Running %r with env %r", self.status_cgi_path, env)
            p = subprocess.run(self.status_cgi_path, env=env, capture_output=True)
            if p.returncode == 0 and p.stdout != "":
                return self._parse_cgi_json(p.stdout)
            LOG.warning(
                "Calling %r failed with %s: %r, %r",
                self.status_cgi_path,
                p.returncode,
                p.stdout[:255],
                p.stderr[:255],
            )
            time.sleep(1)

        return None

    def _annotation_name_from_url(self, url):
        if self.notes_map is None:
            return "link"
        for pattern, name in self.notes_map.items():
            if pattern.search(url):
                return name
        return "link"

    def _annotations_from_notes_url(self, problem):
        notes_url = problem.get("notes_url")
        if notes_url is None:
            return {}

        if "' '" in notes_url:
            urls = re.split("' '", notes_url.strip("'"))
        else:
            urls = [notes_url]

        res = {}
        for raw_url in urls:
            url = self._quote_url(raw_url)
            key = self._annotation_name_from_url(url)
            if key not in res:
                res[key] = url
            else:
                for i in range(1, 5):
                    newkey = key + str(i)
                    if newkey not in res:
                        res[newkey] = url
                        break
        return res

    def _quote_url(self, url):
        """Quote URL components to form a valid URL (e.g. query string parts are url-encoded)"""

        # Split components first
        url_parts = urllib.parse.urlsplit(url)
        qs_parts = urllib.parse.parse_qsl(url_parts.query)
        qs_encoded = urllib.parse.urlencode(qs_parts)

        # Reassemble parts by replacing 'query' with encoded counterpart
        parts_dict = url_parts._asdict()
        del parts_dict["query"]

        # Assemble
        url_parts_quoted = urllib.parse.SplitResult(query=qs_encoded, **parts_dict)
        return urllib.parse.urlunsplit(url_parts_quoted)

    def alerts_from_status_json(self, status):
        res = []
        # XXX parse host problems
        for problem in status["status"]["service_status"]:
            host_name = problem["host_name"]
            service = problem["service_description"]
            plugin_output = problem["status_information"]
            severity = problem["status"]
            if "#page" in service and severity.lower() == "critical":
                severity = "page"

            labels = {
                "alertname": service,
                "severity": severity.lower(),
                "instance": "%s" % host_name,
                "source": "icinga",
            }
            if self.team_tag_matcher is not None:
                team_tag = self.team_tag_matcher.get_matching_team(
                    alertname=labels["alertname"], instance=labels["instance"]
                )
                if team_tag:
                    labels["team"] = team_tag

            annotations = {"summary": plugin_output}
            annotations.update(self._annotations_from_notes_url(problem))

            generator_url = None
            if self.extinfo_url is not None:
                extinfo_qs = [
                    "type=2",
                    "host=" + host_name,
                    "service=" + service,
                ]

                generator_url = self.extinfo_url + "?" + "&".join(extinfo_qs)
                generator_url = self._quote_url(generator_url)

            res.append(
                am_client.PostableAlert(
                    labels=labels, annotations=annotations, generator_url=generator_url
                )
            )

        return res

    @parse_duration.time()
    def _get_current_alerts(self):
        """Transform Icinga' status JSON into AM's PostableAlert(s)"""

        res = []
        status_json = self._json_from_status_cmd()
        if not status_json:
            self.metrics["parse_errors_total"].inc()
            return res

        try:
            res = self.alerts_from_status_json(status_json)
            self.metrics["last_parse_success_epoch"].set(time.time())
        except KeyError as e:
            LOG.exception(e)
            return []

        return res

    def _alerts_dict(self, alerts):
        """Return a dictionary with alert labels as keys, and the alert as value.

        Labels are used to uniquely identify alerts."""

        res = {}
        for alert in alerts:
            res[frozenset(alert.labels.items())] = alert
        return res

    def _resolved_alerts(self, current, last):
        """Return the list of PostableAlert that were resolved since last run."""
        now = datetime.datetime.now().astimezone()  # ends_at needs to have a tz
        res = []

        unique_last = self._alerts_dict(last)
        unique_current = self._alerts_dict(current)

        for labels, alert in unique_last.items():
            if labels in unique_current:
                continue
            # Copy the PostableAlert and pretend it ended "now"
            resolved = am_client.PostableAlert(**alert.to_dict())
            resolved.ends_at = now
            res.append(resolved)

        return res

    def _get_alerts(self):
        """Return the list of PostableAlert(s) to be sent to AM.

        Including current alerts and previously-resolved alerts."""
        res = []

        current_alerts = self._get_current_alerts()

        if self.alerts is None:
            resolved_alerts = []
        else:
            resolved_alerts = self._resolved_alerts(current_alerts, self.alerts)

        res = resolved_alerts + current_alerts

        # Set alerts to current for next iteration
        self.alerts = current_alerts

        return res

    def _send_alerts(self):
        alerts = self._get_alerts()

        for client in self.am_clients:
            try:
                LOG.debug(
                    "Sending %r alerts to %r",
                    len(alerts),
                    client.api_client.configuration.host,
                )
                client.post_alerts(alerts, _request_timeout=AM_REQUEST_TIMEOUT)
                self.metrics["send_completed_total"].inc()
            except (
                urllib3.exceptions.HTTPError,
                am_client.exceptions.ApiException,
            ) as e:
                LOG.exception(e)
                self.metrics["send_errors_total"].inc()
                continue


def main(args: Optional[List[str]] = None):
    """Main"""
    args = sys.argv[1:] if args is None else args

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-l",
        "--listen",
        metavar="ADDRESS",
        help="Listen on this address",
        default=":9247",
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="Enable debug logging"
    )
    parser.add_argument(
        "-u",
        "--status.querystring",
        metavar="STRING",
        help="The query string for status.cgi",
        default="allunhandledproblems&serviceprops=270336&hostprops=270336&jsonoutput",
        dest="query_string",
    )
    parser.add_argument(
        "--status.path",
        metavar="PATH",
        help="The path to status.cgi",
        default="/usr/lib/cgi-bin/icinga/status.cgi",
        dest="status_cgi_path",
    )
    parser.add_argument(
        "--interval",
        metavar="SECONDS",
        help="Send alerts every SECONDS",
        default=10,
        type=int,
    )
    parser.add_argument(
        "-a",
        "--alertmanager.url",
        help="Alertmanager(s) URL. Defaults to http://localhost:9093",
        type=str,
        metavar="URL",
        dest="alertmanagers",
        action="append",
        default=None,
    )
    parser.add_argument(
        "--labels.team.config-file",
        dest="labels_team_config_file",
        help=(
            "Path to the config file with the team definition. It should be a YAML formatted file "
            'with a single hash mapping in the form \'{"team_name1": {"alertname": '
            '[".*regex1.*", ...], "instance": [".*regex2.*", ...]}, "team_name2": ...}\', '
            "where the `alertname` and `instance` keys are literal but optional, and `team_nameN` "
            "and `regexN` are placeholders for the real team name and the actual regexes to match."
            " Note that it will use the first team that matches, so the order is relevant."
        ),
    )
    parser.add_argument(
        "--icinga.extinfo-url",
        dest="extinfo_url",
        default=None,
        help=(
            "The url to Icinga's extinfo.cgi. Used to link alerts in"
            "Alertmanager to their Icinga web counterpart."
        ),
    )
    args = parser.parse_args(args=args)

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARNING)

    if args.alertmanagers is None:
        args.alertmanagers = ["http://localhost:9093"]

    address, port = args.listen.split(":", 1)

    LOG.info("Starting Icinga AM on %s:%s", address, port)

    # XXX make this configurable
    notes_map = {
        re.compile("://grafana"): "dashboard",
        re.compile("://wikitech"): "runbook",
        re.compile("://logstash"): "logs",
        re.compile("://phabricator"): "phab",
        re.compile("://netbox"): "netbox",
    }

    if args.labels_team_config_file:
        team_tag_matcher = TeamTagMatcher.from_team_regexes_file(
            team_regexes_file=open(args.labels_team_config_file, "r")
        )
    else:
        team_tag_matcher = None

    icinga_am = PrometheusIcingaAM(
        args.alertmanagers,
        args.query_string,
        args.status_cgi_path,
        args.interval,
        notes_map=notes_map,
        team_tag_matcher=team_tag_matcher,
        extinfo_url=args.extinfo_url,
    )
    icinga_am.start()

    start_http_server(int(port), addr=address)
    # In lieu of a infinite loop, join the thread started by exposition.start_http_server
    try:
        threading.enumerate()[-1].join()
    except KeyboardInterrupt:
        return 1

    return 0


if __name__ == "__main__":
    exit(main())
