#!/usr/bin/env python3
# Copyright 2018 Cole White
#                Wikimedia Foundation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Prometheus Icinga Exporter"""

import argparse
import logging
import os
import threading
from time import sleep

from prometheus_client import start_http_server
from prometheus_client.core import REGISTRY, CounterMetricFamily, GaugeMetricFamily, Summary

LOG = logging.getLogger(__name__)


class PrometheusIcingaExporter:
    """Prometheus Icinga Exporter"""

    scrape_duration = Summary("icinga_scrape_duration_seconds", "Icinga exporter scrape duration", ())

    def __init__(self, status_file=""):
        self.status_file = status_file
        self.check_execution_times = []
        self.check_latencies = []
        self.host_count = 0
        self.service_count = 0
        self.status_mtime = 0
        self.work_units = 0.0
        self.is_source_file_complete = False
        self.initial_run = True
        self.file_not_found = False

    @classmethod
    def get_block_type_line_and_content(cls, lines):
        """
        Generator that yields the block data in list form and the line the block begins at.

        :param lines: raw status.dat data
        :return: tuple (block_type, block_begins_at_line, block_data)
        """
        block_begins_at_line = 0
        current_line = 0
        block_data = []
        in_block = False
        block_type = None
        for line in lines.split("\n"):
            current_line += 1
            line = line.strip()
            if not line:
                continue
            if line[0] == "#":
                continue
            if line == "}":
                if in_block:
                    yield block_type, block_begins_at_line, block_data
                in_block = False
                block_type = None
                block_data = []
            if in_block:
                block_data.append(line)
            if line == "hoststatus {":
                in_block = True
                block_begins_at_line = current_line
                block_type = "host"
            if line == "servicestatus {":
                in_block = True
                block_begins_at_line = current_line
                block_type = "service"
            if line == "contactstatus {":
                in_block = True
                block_begins_at_line = current_line
                block_type = "contact"

    def load_stats(self):
        """Reads state from status_file and populates metrics"""
        self.check_latencies = []
        self.check_execution_times = []
        self.host_count = 0
        self.service_count = 0
        self.status_mtime = 0
        self.work_units = 0.0

        if not os.path.isfile(self.status_file):
            status_file = autodiscover_status_file()
        else:
            status_file = self.status_file

        self.status_mtime = get_mod_time(status_file)

        with open(status_file, "r", encoding="ISO-8859-1") as file_handle:
            raw_status_dat = file_handle.read()

        for (
            block_type,
            block_begins_at_line,
            block_data,
        ) in self.get_block_type_line_and_content(raw_status_dat):
            if block_type == "contact":  # !!!Warning total hack right here!!!
                # contact blocks are at the end of status.dat.  If present, we can say we read all we needed.
                self.is_source_file_complete = True
                break
            if block_type == "host":
                self.host_count += 1
            if block_type == "service":
                self.service_count += 1
            try:
                self.check_latencies.append(float(get_attribute(block_data, "check_latency")))
                self.check_execution_times.append(float(get_attribute(block_data, "check_execution_time")))
                self.work_units += 1.0 / float(get_attribute(block_data, "check_interval"))
            except KeyError as ex:  # handle key not found errors
                LOG.error("%s in block at line %s.", ex, block_begins_at_line)
            except ValueError as ex:  # handle float conversion errors
                LOG.error("%s in block at line %s", ex, block_begins_at_line)

    @scrape_duration.time()
    def collect(self):
        """
        The function that prometheus_client expects to get metrics from.

        :return: Metric Family
        """
        # Prometheus client appears to run collect() on register().
        # If retry logic raises the RuntimeError on this initial run, the service fails to start.
        # Circumvent this by skipping the initial run.
        if self.initial_run:
            self.initial_run = False
            return
        self.is_source_file_complete = False
        # Handle case where file does not exist.
        self.file_not_found = False
        retry_count = 0
        try:
            self.load_stats()
        except FileNotFoundError as e:
            LOG.error("Problem gathering metrics:", exc_info=e)
            self.file_not_found = True
        while not self.is_source_file_complete and not self.file_not_found:
            if retry_count < 3:
                retry_count += 1
                self.load_stats()
            else:
                raise RuntimeError("Retry limit reached.  Status file reads were incomplete.")
            sleep(1)
        if self.file_not_found:
            metrics = [build_status(0)]
        else:
            metrics = [
                build_gauge(
                    "icinga_avg_check_latency",
                    "Average check latency",
                    (sum(self.check_latencies) / len(self.check_latencies)),
                ),
                build_gauge(
                    "icinga_max_check_latency",
                    "Max check latency",
                    max(self.check_latencies),
                ),
                build_gauge(
                    "icinga_min_check_latency",
                    "Min check latency",
                    min(self.check_latencies),
                ),
                build_gauge(
                    "icinga_avg_check_execution_time",
                    "Average check execution time",
                    (sum(self.check_execution_times) / len(self.check_execution_times)),
                ),
                build_gauge(
                    "icinga_max_check_execution_time",
                    "Max check execution time",
                    max(self.check_execution_times),
                ),
                build_gauge(
                    "icinga_min_check_execution_time",
                    "Min check execution time",
                    min(self.check_execution_times),
                ),
                build_gauge("icinga_host_count", "Number of hosts", self.host_count),
                build_gauge("icinga_service_count", "Number of services", self.service_count),
                build_gauge(
                    "icinga_workunits_count",
                    "Sum across all checks of (1/check_interval). "
                    "How many checks-per-second the configuration asks Icinga to execute.",
                    self.work_units,
                ),
                build_gauge(
                    "icinga_status_file_mod_time",
                    "Status file modification time.",
                    self.status_mtime,
                ),
                build_status(1),
            ]

        for metric in metrics:
            yield metric


def build_status(value):
    """
    Build a GaugeMetricFamily instance pre-formatted for scrape status.

    :param value: int 1 indicating success or 0 indicating failure
    :return: GaugeMetricFamily instance
    """
    return build_gauge("icinga_exporter_scrape_success", "Was gathering the metrics successful?", value)


def build_gauge(name, description, value):
    """
    Build a GaugeMetricFamily object and populate it with the value.

    :param name: metric name as expected in prometheus
    :param description: description of metric
    :param value: metric value
    :return: GaugeMetricFamily instance
    """
    gauge = GaugeMetricFamily(name, description)
    gauge.add_metric([], value)
    return gauge


def build_counter(name, description, value):
    """
    Build a CounterMetricFamily object and populate it with the value.

    :param name: metric name as expected in prometheus
    :param description: description of metric
    :param value: metric value
    :return: CounterMetricFamily instance
    """
    counter = CounterMetricFamily(name, description)
    counter.add_metric([], value)
    return counter


def extract_status_file_path(lines):
    """
    Slices each line looking for status_file and returns the value
    :param lines: string
    :return: string
    """
    for line in lines.split("\n"):
        if line[0:11] == "status_file":
            return get_key_and_value(line)[1].strip()
    raise KeyError("Could not locate status_file option!")


def get_key_and_value(line):
    """
    Extract key and value from line.

    :param line: line in status block
    :return: (str(key), str(value))
    """
    idx = 0
    for char in line:
        if char == "=":
            break
        idx += 1
    key = line[0:idx]
    value = line[idx + 1 :]
    return key, value


def get_attribute(lines, attribute):
    """
    Searches block for attribute and returns the value if it is found.
    Otherwise raises KeyError.

    :param lines: list of lines constituing a block
    :param attribute: the attribute key
    :return: string(value)
    """
    for line in lines:
        key, value = get_key_and_value(line)
        if key == attribute:
            return value
    raise KeyError("Unable to find attribute {}".format(attribute))


def autodiscover_status_file():
    """
    Attempts to discover the status file via /etc/icinga/icinga.cfg
    Searches for the status_file key and returns the value
    :return: string the status_file path
    """
    icinga_cfg_path = "/etc/icinga/icinga.cfg"
    if os.path.isfile(icinga_cfg_path):
        with open(icinga_cfg_path, "r") as config_file:
            return extract_status_file_path(config_file.read())
    raise FileNotFoundError("Unable to load icinga config: {}".format(icinga_cfg_path))


def get_mod_time(status_file):
    return os.stat(status_file).st_mtime


def main():
    """Main"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-l",
        "--listen",
        metavar="ADDRESS",
        help="Listen on this address",
        default=":9245",
    )
    parser.add_argument("-d", "--debug", action="store_true", help="Enable debug logging")
    parser.add_argument(
        "-f",
        "--file",
        metavar="FILE",
        help="Fetch stats from this file",
        default="./status.dat",
    )
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARNING)

    address, port = args.listen.split(":", 1)

    LOG.info("Starting icinga_exporter on %s:%s", address, port)

    REGISTRY.register(PrometheusIcingaExporter(args.file))
    start_http_server(int(port), addr=address)
    # In lieu of a infinite loop, join the thread started by exposition.start_http_server
    try:
        threading.enumerate()[-1].join()
    except KeyboardInterrupt:
        return 1


if __name__ == "__main__":
    exit(main())
